package com.icing.dialer.modem

import org.apache.commons.math3.complex.Complex
import org.apache.commons.math3.transform.DftNormalization
import org.apache.commons.math3.transform.FastFourierTransformer
import org.apache.commons.math3.transform.TransformType
import kotlin.math.*

class FSKDemodulator {
    private val fft = FastFourierTransformer(DftNormalization.STANDARD)
    private val symbolBuffer = mutableListOf<Int>()
    private var sampleBuffer = FloatArray(0)
    private var syncFound = false
    private var syncPosition = 0
    
    // Moving average filters for each frequency
    private val freq00Filter = MovingAverageFilter(FSKConstants.SAMPLES_PER_SYMBOL / 2)
    private val freq01Filter = MovingAverageFilter(FSKConstants.SAMPLES_PER_SYMBOL / 2)
    private val freq10Filter = MovingAverageFilter(FSKConstants.SAMPLES_PER_SYMBOL / 2)
    private val freq11Filter = MovingAverageFilter(FSKConstants.SAMPLES_PER_SYMBOL / 2)
    
    // Demodulate audio samples to symbols
    fun demodulateSamples(samples: FloatArray): IntArray {
        val symbols = mutableListOf<Int>()
        
        // Process samples in chunks of SAMPLES_PER_SYMBOL
        for (i in 0 until samples.size - FSKConstants.SAMPLES_PER_SYMBOL step FSKConstants.SAMPLES_PER_SYMBOL) {
            val symbolSamples = samples.sliceArray(i until i + FSKConstants.SAMPLES_PER_SYMBOL)
            val symbol = detectSymbol(symbolSamples)
            symbols.add(symbol)
        }
        
        return symbols.toIntArray()
    }
    
    // Non-coherent detection using Goertzel algorithm for efficiency
    private fun detectSymbol(samples: FloatArray): Int {
        val power00 = goertzelMagnitude(samples, FSKConstants.FREQ_00)
        val power01 = goertzelMagnitude(samples, FSKConstants.FREQ_01)
        val power10 = goertzelMagnitude(samples, FSKConstants.FREQ_10)
        val power11 = goertzelMagnitude(samples, FSKConstants.FREQ_11)
        
        // Apply moving average filter to reduce noise
        val filtered00 = freq00Filter.filter(power00)
        val filtered01 = freq01Filter.filter(power01)
        val filtered10 = freq10Filter.filter(power10)
        val filtered11 = freq11Filter.filter(power11)
        
        // Find maximum power
        val powers = floatArrayOf(filtered00, filtered01, filtered10, filtered11)
        var maxIndex = 0
        var maxPower = powers[0]
        
        for (i in 1 until powers.size) {
            if (powers[i] > maxPower) {
                maxPower = powers[i]
                maxIndex = i
            }
        }
        
        return maxIndex
    }
    
    // Goertzel algorithm for single frequency detection
    private fun goertzelMagnitude(samples: FloatArray, targetFreq: Double): Float {
        val k = round(samples.size * targetFreq / FSKConstants.SAMPLE_RATE).toInt()
        val omega = 2.0 * PI * k / samples.size
        val cosine = cos(omega)
        val coeff = 2.0 * cosine
        
        var q0 = 0.0
        var q1 = 0.0
        var q2 = 0.0
        
        for (sample in samples) {
            q0 = coeff * q1 - q2 + sample
            q2 = q1
            q1 = q0
        }
        
        val real = q1 - q2 * cosine
        val imag = q2 * sin(omega)
        
        return sqrt(real * real + imag * imag).toFloat()
    }
    
    // Find preamble in audio stream
    fun findPreamble(samples: FloatArray): Int {
        val preamblePattern = intArrayOf(1, 2, 1, 2, 1, 2, 1, 2) // 01 10 01 10...
        val correlationThreshold = 0.8f
        
        for (i in 0 until samples.size - (preamblePattern.size * FSKConstants.SAMPLES_PER_SYMBOL)) {
            var correlation = 0.0f
            var patternPower = 0.0f
            var signalPower = 0.0f
            
            for (j in preamblePattern.indices) {
                val startIdx = i + j * FSKConstants.SAMPLES_PER_SYMBOL
                val endIdx = startIdx + FSKConstants.SAMPLES_PER_SYMBOL
                
                if (endIdx <= samples.size) {
                    val symbolSamples = samples.sliceArray(startIdx until endIdx)
                    val detectedSymbol = detectSymbol(symbolSamples)
                    
                    if (detectedSymbol == preamblePattern[j]) {
                        correlation += 1.0f
                    }
                    
                    // Calculate signal power for SNR estimation
                    for (sample in symbolSamples) {
                        signalPower += sample * sample
                    }
                }
            }
            
            val normalizedCorrelation = correlation / preamblePattern.size
            if (normalizedCorrelation >= correlationThreshold) {
                return i
            }
        }
        
        return -1 // Preamble not found
    }
    
    // Convert symbols back to bytes
    fun symbolsToBytes(symbols: IntArray): ByteArray {
        val bytes = ByteArray(symbols.size / 4)
        var byteIndex = 0
        
        for (i in symbols.indices step 4) {
            if (i + 3 < symbols.size) {
                val byte = ((symbols[i] and 0x03) shl 6) or
                          ((symbols[i + 1] and 0x03) shl 4) or
                          ((symbols[i + 2] and 0x03) shl 2) or
                          (symbols[i + 3] and 0x03)
                bytes[byteIndex++] = byte.toByte()
            }
        }
        
        return bytes.sliceArray(0 until byteIndex)
    }
    
    // Carrier frequency offset estimation and correction
    fun estimateFrequencyOffset(samples: FloatArray): Double {
        // Use pilot tone or known preamble for frequency offset estimation
        val fftSize = 1024
        val paddedSamples = samples.copyOf(fftSize)
        
        // Convert to complex array for FFT
        val complexSamples = Array(fftSize) { i ->
            if (i < samples.size) Complex(paddedSamples[i].toDouble()) else Complex.ZERO
        }
        
        val spectrum = fft.transform(complexSamples, TransformType.FORWARD)
        
        // Find peak frequencies
        var maxMagnitude = 0.0
        var peakBin = 0
        
        for (i in spectrum.indices) {
            val magnitude = spectrum[i].abs()
            if (magnitude > maxMagnitude) {
                maxMagnitude = magnitude
                peakBin = i
            }
        }
        
        // Calculate frequency offset
        val detectedFreq = peakBin * FSKConstants.SAMPLE_RATE.toDouble() / fftSize
        val expectedFreq = (FSKConstants.FREQ_00 + FSKConstants.FREQ_11) / 2 // Center frequency
        
        return detectedFreq - expectedFreq
    }
    
    // Reset demodulator state
    fun reset() {
        symbolBuffer.clear()
        sampleBuffer = FloatArray(0)
        syncFound = false
        syncPosition = 0
        freq00Filter.reset()
        freq01Filter.reset()
        freq10Filter.reset()
        freq11Filter.reset()
    }
    
    // Simple moving average filter
    private class MovingAverageFilter(private val windowSize: Int) {
        private val buffer = FloatArray(windowSize)
        private var index = 0
        private var sum = 0.0f
        private var count = 0
        
        fun filter(value: Float): Float {
            sum -= buffer[index]
            buffer[index] = value
            sum += value
            index = (index + 1) % windowSize
            
            if (count < windowSize) {
                count++
            }
            
            return sum / count
        }
        
        fun reset() {
            buffer.fill(0.0f)
            index = 0
            sum = 0.0f
            count = 0
        }
    }
}