package com.icing.dialer.modem

import android.media.*
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.flow.*
import java.util.concurrent.ConcurrentLinkedQueue

class FSKModem {
    private val audioCodec = AudioCodec()
    private val modulator = FSKModulator()
    private val demodulator = FSKDemodulator()
    private val frameProcessor = FrameProcessor()
    
    private var audioRecord: AudioRecord? = null
    private var audioTrack: AudioTrack? = null
    
    private val txQueue = ConcurrentLinkedQueue<ByteArray>()
    private val rxQueue = ConcurrentLinkedQueue<ByteArray>()
    
    private val scope = CoroutineScope(Dispatchers.IO + SupervisorJob())
    private var isRunning = false
    
    // Flow for received data
    private val _receivedData = MutableSharedFlow<ByteArray>()
    val receivedData: SharedFlow<ByteArray> = _receivedData.asSharedFlow()
    
    // Modem states
    enum class ModemState {
        IDLE, TRANSMITTING, RECEIVING, ERROR
    }
    
    private val _state = MutableStateFlow(ModemState.IDLE)
    val state: StateFlow<ModemState> = _state.asStateFlow()
    
    fun initialize() {
        setupAudioRecord()
        setupAudioTrack()
    }
    
    private fun setupAudioRecord() {
        val bufferSize = AudioRecord.getMinBufferSize(
            FSKConstants.SAMPLE_RATE,
            AudioFormat.CHANNEL_IN_MONO,
            AudioFormat.ENCODING_PCM_16BIT
        )
        
        audioRecord = AudioRecord(
            MediaRecorder.AudioSource.MIC,
            FSKConstants.SAMPLE_RATE,
            AudioFormat.CHANNEL_IN_MONO,
            AudioFormat.ENCODING_PCM_16BIT,
            bufferSize * 2
        )
    }
    
    private fun setupAudioTrack() {
        val bufferSize = AudioTrack.getMinBufferSize(
            FSKConstants.SAMPLE_RATE,
            AudioFormat.CHANNEL_OUT_MONO,
            AudioFormat.ENCODING_PCM_16BIT
        )
        
        audioTrack = AudioTrack(
            AudioManager.STREAM_MUSIC,
            FSKConstants.SAMPLE_RATE,
            AudioFormat.CHANNEL_OUT_MONO,
            AudioFormat.ENCODING_PCM_16BIT,
            bufferSize * 2,
            AudioTrack.MODE_STREAM
        )
    }
    
    fun start() {
        if (isRunning) return
        
        isRunning = true
        audioRecord?.startRecording()
        audioTrack?.play()
        
        // Start coroutines for TX and RX
        scope.launch { transmitLoop() }
        scope.launch { receiveLoop() }
    }
    
    fun stop() {
        isRunning = false
        audioRecord?.stop()
        audioTrack?.stop()
        scope.cancel()
    }
    
    fun sendData(data: ByteArray) {
        txQueue.offer(data)
    }
    
    private suspend fun transmitLoop() {
        val audioBuffer = ShortArray(FSKConstants.OPUS_FRAME_SIZE)
        
        while (isRunning) {
            if (txQueue.isNotEmpty()) {
                _state.value = ModemState.TRANSMITTING
                
                val data = txQueue.poll() ?: continue
                
                try {
                    // Encode audio data with Opus
                    val encodedAudio = audioCodec.encode(data.toShortArray())
                    
                    // Create frame with error correction
                    val frame = frameProcessor.createFrame(encodedAudio)
                    val frameBytes = frame.toByteArray()
                    
                    // Convert to symbols
                    val symbols = modulator.bytesToSymbols(frameBytes)
                    
                    // Generate preamble
                    val preamble = modulator.generatePreamble()
                    
                    // Modulate symbols
                    val modulatedData = modulator.modulateSymbols(symbols)
                    
                    // Apply raised cosine filter
                    val filtered = modulator.applyRaisedCosineFilter(modulatedData)
                    
                    // Combine preamble and data
                    val txSamples = FloatArray(preamble.size + filtered.size)
                    System.arraycopy(preamble, 0, txSamples, 0, preamble.size)
                    System.arraycopy(filtered, 0, txSamples, preamble.size, filtered.size)
                    
                    // Convert to 16-bit PCM and transmit
                    val pcmData = ShortArray(txSamples.size) { i ->
                        (txSamples[i] * 32767).toInt().coerceIn(-32768, 32767).toShort()
                    }
                    
                    audioTrack?.write(pcmData, 0, pcmData.size)
                    
                } catch (e: Exception) {
                    _state.value = ModemState.ERROR
                    e.printStackTrace()
                }
                
                _state.value = ModemState.IDLE
            }
            
            delay(10) // Small delay to prevent busy waiting
        }
    }
    
    private suspend fun receiveLoop() {
        val audioBuffer = ShortArray(FSKConstants.AUDIO_BUFFER_SIZE)
        val sampleBuffer = mutableListOf<Float>()
        
        while (isRunning) {
            val bytesRead = audioRecord?.read(audioBuffer, 0, audioBuffer.size) ?: 0
            
            if (bytesRead > 0) {
                _state.value = ModemState.RECEIVING
                
                // Convert to float samples
                val samples = FloatArray(bytesRead) { i ->
                    audioBuffer[i] / 32768.0f
                }
                
                sampleBuffer.addAll(samples.toList())
                
                // Look for preamble
                if (sampleBuffer.size >= FSKConstants.PREAMBLE_LENGTH * FSKConstants.SAMPLES_PER_SYMBOL) {
                    val bufferArray = sampleBuffer.toFloatArray()
                    val preambleIndex = demodulator.findPreamble(bufferArray)
                    
                    if (preambleIndex >= 0) {
                        // Preamble found, extract frame
                        val frameStart = preambleIndex + 
                            (FSKConstants.PREAMBLE_LENGTH * FSKConstants.SAMPLES_PER_SYMBOL)
                        
                        if (frameStart < bufferArray.size) {
                            // Estimate and correct frequency offset
                            val frameSection = bufferArray.sliceArray(
                                frameStart until minOf(
                                    frameStart + FSKConstants.FRAME_SIZE * 4 * FSKConstants.SAMPLES_PER_SYMBOL,
                                    bufferArray.size
                                )
                            )
                            
                            // Demodulate symbols
                            val symbols = demodulator.demodulateSamples(frameSection)
                            
                            // Convert symbols to bytes
                            val frameBytes = demodulator.symbolsToBytes(symbols)
                            
                            // Process frame (error correction and CRC check)
                            val decodedData = frameProcessor.processFrame(frameBytes)
                            
                            if (decodedData != null) {
                                // Decode audio with Opus
                                val audioData = audioCodec.decode(decodedData)
                                
                                // Emit received data
                                _receivedData.emit(audioData.toByteArray())
                            }
                            
                            // Remove processed samples
                            sampleBuffer.subList(0, frameStart + frameSection.size).clear()
                        }
                    }
                    
                    // Limit buffer size to prevent memory issues
                    if (sampleBuffer.size > FSKConstants.SAMPLE_RATE * 2) {
                        sampleBuffer.subList(0, FSKConstants.SAMPLE_RATE).clear()
                    }
                }
                
                _state.value = ModemState.IDLE
            }
            
            delay(10)
        }
    }
    
    fun release() {
        stop()
        audioRecord?.release()
        audioTrack?.release()
        audioCodec.release()
        audioRecord = null
        audioTrack = null
    }
    
    // Utility extension functions
    private fun ByteArray.toShortArray(): ShortArray {
        return ShortArray(size / 2) { i ->
            ((this[i * 2].toInt() and 0xFF) or 
             ((this[i * 2 + 1].toInt() and 0xFF) shl 8)).toShort()
        }
    }
    
    private fun ShortArray.toByteArray(): ByteArray {
        val bytes = ByteArray(size * 2)
        for (i in indices) {
            bytes[i * 2] = (this[i].toInt() and 0xFF).toByte()
            bytes[i * 2 + 1] = ((this[i].toInt() shr 8) and 0xFF).toByte()
        }
        return bytes
    }
}